---@type Tinkr
local Tinkr,
---@class Bastion
Bastion = ...

-- Create a module class for a bastion module

---@class Bastion.Module
---@field name string
---@field enabled boolean
---@field synced function[]
---@field interval? number
---@field nextTick number
local Module = {}
Module.__index = Module

-- __tostring
---@return string
function Module:__tostring()
    return "Bastion.__Module(" .. self.name .. ")"
end

-- Constructor
---@param name string
---@param interval? number
---@return Bastion.Module
function Module:New(name, interval)
    local module = {}
    setmetatable(module, Module)

    module.name = name
    module.enabled = false
    module.synced = {}
    module.interval = interval
    module.nextTick = 0
    return module
end

-- Enable the module
function Module:Enable()
    self.enabled = true
    return self
end

-- Disable the module
function Module:Disable()
    self.enabled = false
    return self
end

-- Toggle the module
---@return nil
function Module:Toggle()
    if self.enabled then
        self:Disable()
    else
        self:Enable()
    end
    return self
end

-- Add a function to the sync list
---@param func function
function Module:Sync(func)
    table.insert(self.synced, func)
    return self
end

-- Remove a function from the sync list
---@param func function
function Module:Unsync(func)
    for i = 1, #self.synced do
        if self.synced[i] == func then
            table.remove(self.synced, i)
            break
        end
    end
    return self
end

-- Sync
function Module:Tick()
    if self.enabled then
        if Bastion.Tick >= self.nextTick then
            self.nextTick = Bastion.Tick + (self.interval or 1)
            for i = 1, #self.synced do
                self.synced[i]()
            end
        end
    end
    return self
end

Bastion.Module = Module

---@type Bastion.Module[]
local MODULES = {}

function Bastion:TickModules()
    for i = 1, #MODULES do
        MODULES[i]:Tick()
    end
end

-- Find a module by name
---@param name string
function Bastion:FindModule(name)
    for i = 1, #MODULES do
        if MODULES[i].name == name then
            return MODULES[i]
        end
    end
    return false
end

---@param module Bastion.Module
function Bastion:Register(module)
    local foundModule = Bastion:FindModule(module.name)
    if not foundModule then
        table.insert(MODULES, module)
        Bastion.Util:Print("Registered", module)
    else
        Bastion.Util:Print("Module already registered", module)
        return
    end
end

---@param args { [2]: string }
Bastion.Globals.Command:Register("module", "Toggle a module on/off", function(args)
    local module = Bastion:FindModule(args[2])
    if module then
        module:Toggle()
        if module.enabled then
            Bastion.Util:Print("Enabled", module.name)
        else
            Bastion.Util:Print("Disabled", module.name)
        end
    else
        Bastion.Util:Print("Module not found")
    end
end)
