---@type Tinkr
local Tinkr,
---@class Bastion
Bastion = ...

---@class Bastion.Spell.Traits.Cast
---@field moving boolean
---@field dead boolean
---@field global boolean
---@field casting boolean
---@field channeling boolean
---@field override boolean
---@field talent boolean | spellId
---@field power boolean

---@class Bastion.Spell.Traits.Cost
---@field type Enum.PowerType
---@field cost number
---@field minCost number
---@field requiredAuraID number
---@field costPercent number
---@field costPerSecond number

---@class Bastion.Spell.Traits.Target
---@field exists boolean
---@field player boolean

---@class Bastion.Spell.Traits
---@field cast Bastion.Spell.Traits.Cast
---@field target Bastion.Spell.Traits.Target
---@field cost Bastion.Spell.Traits.Cost[]

---@class Bastion.Spell.Traits.Cast.Params : Bastion.Spell.Traits.Cast, { [string]?: boolean }
---@class Spell.Traits.Target.Params : Bastion.Spell.Traits.Target, { [string]?: boolean }
---@class Spell.Traits.Cost.Params : Bastion.Spell.Traits.Cost[]

---@class Bastion.Spell.Traits.Params
---@field cast? Bastion.Spell.Traits.Cast.Params
---@field target? Spell.Traits.Target.Params
---@field cost? Spell.Traits.Cost.Params

---@class Bastion.Spell.Aura
---@field spell Bastion.Spell
---@field source? Bastion.Unit
---@field target? Bastion.Unit
---@field lastApplied number

-- Create a new Spell class
---@class Bastion.Spell
---@field auras table<spellId, Bastion.Spell.Aura>
---@field CastableIfFunc false | fun(self:Bastion.Spell):boolean
---@field conditions { [string]: { func: fun(self:Bastion.Spell):boolean } }
---@field damage number
---@field damageFormula false | fun(self:Bastion.Spell):number
---@field lastCastAt number | false
---@field lastCastAttempt number | false
---@field OnCastFunc fun(self:Bastion.Spell) | false
---@field overrides { [spellId]: fun(self: Bastion.Spell): Bastion.Spell }
---@field PostCastFunc fun(self:Bastion.Spell) | false
---@field PreCastFunc fun(self:Bastion.Spell) | false
---@field release_at false
---@field spellID number
---@field target Bastion.Unit | false
---@field traits Bastion.Spell.Traits
---@field wasLooking boolean
local Spell = {
    CastableIfFunc = false,
    damage = 0,
    damageFormula = false,
    lastCastAt = false,
    lastCastAttempt = false,
    OnCastFunc = false,
    PostCastFunc = false,
    PreCastFunc = false,
    release_at = false,
}

local usableExcludes = {
    [18562] = true,
}

function Spell:__index(k)
    local response = Bastion.ClassMagic:Resolve(Spell, k)

    if response == nil then
        response = rawget(self, k)
    end

    if response == nil then
        error("Spell:__index: " .. k .. " does not exist")
    end

    return response
end

-- Equals
---@param other Bastion.Spell
function Spell:__eq(other)
    return self:GetID() == other:GetID()
end

-- tostring
function Spell:__tostring()
    return "Bastion.__Spell(" .. self:GetID() .. ")" .. " - " .. self:GetName()
end

-- Constructor
---@param id number
function Spell:New(id)
    ---@class Bastion.Spell
    local self = setmetatable({}, Spell)
    self.auras = {}
    self.overrides = {}
    self.conditions = {}
    self.spellID = id
    self.traits = {
        cast = {
            moving = true,
            dead = false,
            global = false,
            casting = false,
            channeling = false,
            override = false,
            talent = false,
            power = false,
        },
        target = {
            exists = true,
            player = false,
        },
        cost = {},
    }
    self.target = false
    self.wasLooking = false

    return self
end

function Spell:GetDescription()
    return GetSpellDescription(self:GetID()) or ""
end

-- Duplicator
function Spell:Fresh()
    return Spell:New(self:GetID())
end

-- Get the spells id
---@param ignoreOverride? boolean
---@return number
function Spell:GetID(ignoreOverride)
    return ignoreOverride and self.spellID or self:IsOverridden() and self:OverrideSpellID() or self.spellID
end

-- Add post cast func
---@param func fun(self:Bastion.Spell)
function Spell:PostCast(func)
    self.PostCastFunc = func
    return self
end

-- Get the spells name
---@return string
function Spell:GetName()
    return Bastion.Globals.SpellName[self:GetID()] or select(1, GetSpellInfo(self:GetID()))
end

-- Get the spells icon
---@return number
function Spell:GetIcon()
    return select(3, GetSpellInfo(self:GetID()))
end

-- Get the spells cooldown
---@param byId? boolean
---@return number
function Spell:GetCooldown(byId)
    return select(2, GetSpellCooldown(byId and self:GetID() or self:GetName()))
end

-- Return the castable function
function Spell:GetCastableFunction()
    return self.CastableIfFunc
end

-- Return the precast function
function Spell:GetPreCastFunction()
    return self.PreCastFunc
end

-- Get the on cast func
function Spell:GetOnCastFunction()
    return self.OnCastFunc
end

-- Get the spells cooldown remaining
---@param byId? boolean
---@return number
function Spell:GetCooldownRemaining(byId)
    local start, duration = GetSpellCooldown(byId and self:GetID() or self:GetName())
    if start == 0 then
        return 0
    end
    return start + duration - GetTime()
end

-- Get the spell count
---@param byId? boolean
---@return number
function Spell:GetCount(byId)
    return GetSpellCount((byId ~= nil and byId) and self:GetID() or self:GetName())
end

-- On cooldown
---@return boolean
function Spell:OnCooldown()
    return self:GetCooldownRemaining() > 0
end

-- Clear castable function
---@return Bastion.Spell
function Spell:ClearCastableFunction()
    self.CastableIfFunc = false
    return self
end

---@param spell Bastion.Spell
---@param func fun(self: Bastion.Spell): boolean
function Spell:AddOverrideSpell(spell, func)
    self.overrides[spell:GetID()] = func
end

-- Cast the spell
---@param unit Bastion.Unit
---@param condition? string | fun(self:Bastion.Spell):boolean
---@return boolean
function Spell:Cast(unit, condition)
    if not self:Castable() then
        return false
    end

    if condition then
        if type(condition) == "string" and not self:EvaluateCondition(condition) then
            return false
        elseif type(condition) == "function" and not condition(self) then
            return false
        end
    end


    -- Call pre cast function
    if self:GetPreCastFunction() then
        self:GetPreCastFunction()(self)
    end

    -- Check if the mouse was looking
    ---@type boolean
    self.wasLooking = IsMouselooking()

    -- if unit:GetOMToken() contains 'nameplate' then we need to use Object wrapper to cast

    local u = unit and unit:GetOMToken() or self.traits.target.player and "none" or "none"
    if type(u) == "string" and string.find(u, "nameplate") then
        ---@diagnostic disable-next-line: cast-local-type
        u = Object(u)
    end

    -- Cast the spell
    CastSpellByName(self:GetName(), u)
    SpellCancelQueuedSpell()

    Bastion.Util:Debug("Casting", self)

    -- Set the last cast time
    self.lastCastAttempt = GetTime()

    -- Call post cast function
    if self:GetOnCastFunction() then
        self:GetOnCastFunction()(self)
    end

    return true
end

-- ForceCast the spell
---@param unit Bastion.Unit
---@return boolean
function Spell:ForceCast(unit)
    -- Check if the mouse was looking
    self.wasLooking = IsMouselooking()

    -- if unit:GetOMToken() contains 'nameplate' then we need to use Object wrapper to cast
    local u = unit and unit:GetOMToken() or self.traits.target.player and "none" or "none"
    if type(u) == "string" and string.find(u, "nameplate") then
        ---@diagnostic disable-next-line: cast-local-type
        u = Object(u)
    end

    -- Cast the spell
    CastSpellByName(self:GetName(), u)
    SpellCancelQueuedSpell()

    Bastion.Util:Debug("Force Casting", self)

    -- Set the last cast time
    self.lastCastAttempt = GetTime()

    return true
end

-- Get post cast func
function Spell:GetPostCastFunction()
    return self.PostCastFunc
end

function Spell:OverrideSpellID()
    return C_SpellBook.GetOverrideSpell(self.spellID)
end

function Spell:IsOverridden()
    return self:OverrideSpellID() ~= self.spellID
end

-- Check if the spell is known
---@param includeOverrides? boolean
---@return boolean
function Spell:IsKnown(includeOverrides)
    local isKnown = includeOverrides and IsSpellKnownOrOverridesKnown(self:GetID()) or IsSpellKnown(self:GetID()) or
        IsPlayerSpell(self:GetID())
    return isKnown
end

-- Check if the spell is on cooldown
---@return boolean
function Spell:IsOnCooldown()
    local spellIDName = self:IsOverridden() and self:GetName() or self:GetID()
    return select(2, GetSpellCooldown(spellIDName)) > 0
end

---@param byId? boolean
---@return boolean, boolean
function Spell:IsUsableSpell(byId)
    return IsUsableSpell(byId and self:GetID() or self:GetName())
end

-- Check if the spell is usable
---@param byId? boolean
---@return boolean
function Spell:IsUsable(byId)
    local usable, noMana = self:IsUsableSpell(byId)
    return usable and not noMana
end

-- Check if the spell is castable
---@param override? boolean
---@return boolean
function Spell:IsKnownAndUsable(override)
    return self:IsKnown(override) and not self:IsOnCooldown() and self:IsUsable(override ~= nil and false or true)
end

---@param traits Bastion.Spell.Traits.Params
function Spell:SetTraits(traits)
    for _, trait in pairs({ "cast", "target" }) do
        if type(traits[trait]) == "table" then
            for k, _ in pairs(self.traits[trait]) do
                if traits[trait][k] ~= nil then
                    self.traits[trait][k] = traits[trait][k]
                end
            end
        end
    end
    if traits.cost then
        self.traits.cost = {}
        for _, cost in ipairs(traits.cost) do
            table.insert(self.traits.cost, cost)
        end
    end
    return self
end

function Spell:EvaluateTraits()
    local player = Bastion.UnitManager:Get("player")

    if self.traits.cast.power and not self:HasPower() then
        return false
    end

    if not self.traits.cast.global and player:GetGCD() > 0 then
        return false
    end

    if not self.traits.cast.moving and player:IsMoving() then
        return false
    end

    if not self.traits.cast.dead and player:IsDead() then
        return false
    end

    if not self.traits.cast.casting and player:IsCasting() then
        return false
    end

    if not self.traits.cast.channeling and player:IsChanneling() then
        return false
    end

    if self.traits.target.exists and not self:TargetExists() then
        return false
    end

    if self.traits.cast.talent and not IsPlayerSpell(type(self.traits.cast.talent) == "number" and tonumber(self.traits.cast.talent) or self:GetID()) then
        return false
    end

    return true
end

-- Check if the spell is castable
function Spell:Castable()
    if not self:EvaluateTraits() then
        return false
    end
    if self:GetCastableFunction() then
        return self:GetCastableFunction()(self) and
            self:IsKnownAndUsable(type(self.traits.cast.override) ~= nil and self.traits.cast.override or nil)
    end

    return self:IsKnownAndUsable(type(self.traits.cast.override) ~= nil and self.traits.cast.override or nil)
end

-- Set a script to check if the spell is castable
---@param func fun(spell:Bastion.Spell):boolean
function Spell:CastableIf(func)
    self.CastableIfFunc = func
    return self
end

-- Set a script to run before the spell has been cast
---@param func fun(spell:Bastion.Spell)
function Spell:PreCast(func)
    self.PreCastFunc = func
    return self
end

-- Set a script to run after the spell has been cast
---@param func fun(spell:Bastion.Spell)
function Spell:OnCast(func)
    self.OnCastFunc = func
    return self
end

-- Get was looking
---@return boolean
function Spell:GetWasLooking()
    return self.wasLooking
end

-- Click the spell
---@param x number|Bastion.Vector3
---@param y? number
---@param z? number
---@return boolean
function Spell:Click(x, y, z)
    if type(x) == "table" then
        z, y, x = x.z, x.y, x.x
    end
    ---@cast x number
    ---@cast y number
    ---@cast z number
    if IsSpellPending() == 64 then
        MouselookStop()
        Click(x, y, z)
        if self:GetWasLooking() then
            MouselookStart()
        end
        return true
    end
    return false
end

-- Check if the spell is castable and cast it
---@param unit Bastion.Unit
function Spell:Call(unit)
    return self:Cast(unit)
end

-- Check if the spell is castable and cast it
function Spell:HasRange()
    return SpellHasRange(self:GetName())
end

-- Get the range of the spell
---@return number
---@return number
function Spell:GetRange()
    local name, rank, icon, castTime, minRange, maxRange, spellID, originalIcon = GetSpellInfo(self:GetID())
    return maxRange, minRange
end

-- Check if the spell is in range of the unit
---@param unit Bastion.Unit
---@return boolean
function Spell:IsInRange(unit)
    local hasRange = self:HasRange()
    local inRange = IsSpellInRange(self:GetName(), unit:GetOMToken())

    if hasRange == false then
        return true
    end

    if inRange == 1 then
        return true
    end

    return Bastion.UnitManager:Get("player"):InMelee(unit)
end

-- Get the last cast time
---@return number | false
function Spell:GetLastCastTime()
    return self.lastCastAt
end

-- Get time since last cast
---@return number
function Spell:GetTimeSinceLastCast()
    if not self:GetLastCastTime() then
        return math.huge
    end
    return GetTime() - self:GetLastCastTime()
end

-- Get the time since the last cast attempt
---@return number
function Spell:GetTimeSinceLastCastAttempt()
    if not self.lastCastAttempt then
        return math.huge
    end
    return GetTime() - self.lastCastAttempt
end

function Spell:GetChargeInfo()
    return GetSpellCharges(self:GetID())
end

-- Get the spells charges
function Spell:GetCharges()
    return select(1, GetSpellCharges(self:GetID()))
end

function Spell:GetMaxCharges()
    return select(2, GetSpellCharges(self:GetID()))
end

---@return number
function Spell:GetCastLength()
    return select(4, GetSpellInfo(self:GetID()))
end

-- Get the full cooldown (time until all charges are available)
---@return number
function Spell:GetFullRechargeTime()
    local charges, maxCharges, _, duration = self:GetChargeInfo()
    if not charges or not maxCharges or charges == maxCharges then
        return 0
    end

    return (maxCharges - self:GetChargesFractional()) * duration
end

function Spell:Recharge()
    local charges, maxCharges, startTime, duration = self:GetChargeInfo()
    if charges == maxCharges then
        return 0
    end

    local recharge = startTime + duration - GetTime()
    return recharge > 0 and recharge or 0
end

-- Get the spells charges plus fractional
---@return number
function Spell:GetChargesFractional()
    local charges, maxCharges, _, duration = GetSpellCharges(self:GetID())

    if charges == maxCharges then
        return charges
    end
    return charges + ((duration - self:Recharge()) / duration)
end

-- Get the spells charges remaining
---@return number
function Spell:GetChargesRemaining()
    local charges, maxCharges, start, duration = GetSpellCharges(self:GetID())
    return charges
end

-- Create a condition for the spell
---@param name string
---@param func fun(self:Bastion.Spell):boolean
---@return Bastion.Spell
function Spell:Condition(name, func)
    self.conditions[name] = {
        func = func,
    }
    return self
end

-- Get a condition for the spell
---@param name string
---@return { func: fun(self: Bastion.Spell): boolean } | nil
function Spell:GetCondition(name)
    local condition = self.conditions[name]
    if condition then
        return condition
    end

    return nil
end

-- Evaluate a condition for the spell
---@param name string
---@return boolean
function Spell:EvaluateCondition(name)
    local condition = self:GetCondition(name)
    if condition then
        return condition.func(self)
    end

    return false
end

-- Check if the spell has a condition
---@param name string
---@return boolean
function Spell:HasCondition(name)
    local condition = self:GetCondition(name)
    if condition then
        return true
    end

    return false
end

-- Set the spells target
---@param unit Bastion.Unit
---@return Bastion.Spell
function Spell:SetTarget(unit)
    self.target = unit
    return self
end

-- Get the spells target
function Spell:GetTarget()
    return self.target
end

function Spell:TargetExists()
    return self:GetTarget() and self:GetTarget():Exists()
end

-- IsEnrageDispel
---@return boolean
function Spell:IsEnrageDispel()
    return ({
        [2908] = true,
    })[self:GetID()]
end

-- IsMagicDispel
---@return boolean
function Spell:IsMagicDispel()
    return ({
        [88423] = true,
    })[self:GetID()]
end

-- IsCurseDispel
---@return boolean
function Spell:IsCurseDispel()
    return ({
        [88423] = true,
    })[self:GetID()]
end

-- IsPoisonDispel
---@return boolean
function Spell:IsPoisonDispel()
    return ({
        [88423] = true,
    })[self:GetID()]
end

-- IsDiseaseDispel
---@return boolean
function Spell:IsDiseaseDispel()
    return ({})[self:GetID()]
end

-- IsSpell
---@param spell Bastion.Spell
---@return boolean
function Spell:IsSpell(spell)
    return self:GetID() == spell:GetID()
end

function Spell:HasPower()
    local costs = GetSpellPowerCost(self:GetID())
    local checked = {}
    local hasPower = #costs > 0 and false or true
    if not hasPower then
        for _, cost in ipairs(costs) do
            if not checked[cost.type] then
                local powerCost = (cost.cost > cost.minCost and cost.minCost or cost.cost)
                if cost.hasRequiredAura or cost.requiredAuraID == 0 then
                    hasPower = powerCost == 0 or Bastion.UnitManager:Get("player"):GetPower(cost.type) >= powerCost
                    checked[cost.type] = true
                    if not hasPower then
                        return false
                    end
                end
            end
        end
    end
    return hasPower
end

-- GetCost
---@return number
function Spell:GetCost()
    local cost = GetSpellPowerCost(self:GetID())
    return cost and cost[1] and cost[1].cost or 0
end

-- IsFree
---@return boolean
function Spell:IsFree()
    return self:GetCost() == 0
end

---@param damageFormula fun(self:Bastion.Spell): number
function Spell:RegisterDamageFormula(damageFormula)
    self.damageFormula = damageFormula
end

---@return number
function Spell:Damage()
    if self.damageFormula then
        return self:damageFormula()
    else
        return self.damage
    end
end

---@param target? Bastion.Unit
---@param source? "any" | Bastion.Unit
function Spell:GetAura(target, source)
    if type(target) == "nil" then
        target = Bastion.UnitManager:Get("player")
    end
    if type(source) == "nil" then
        source = Bastion.UnitManager:Get("player")
    end

    return target:GetAuras():FindFrom(self, source)
end

---@param spell Bastion.Spell
---@param source? Bastion.Unit
---@param target? Bastion.Unit
function Spell:TrackAura(spell, source, target)
    self.auras[spell:GetID()] = {
        spell = spell,
        source = source,
        target = target,
        lastApplied = 0,
    }
end

---@param aura Bastion.Spell
---@param source? Bastion.Unit
---@param target? Bastion.Unit
function Spell:CheckAuraStatus(aura, source, target)
    for id, trackedAura in pairs(self.auras) do
        if aura:GetID() == id then
            return true
        end
    end
    return false
end

---@param spell Bastion.Spell
---@param source? Bastion.Unit
---@param target? Bastion.Unit
function Spell:UpdateAura(spell, source, target)
    if not self.auras[spell:GetID()] then
        self:TrackAura(spell, source, target)
    end

    self.auras[spell:GetID()].lastApplied = GetTime()
end

Bastion.Spell = Spell
